// VectorTest.cpp - - demo/test routines for BTL_Vector member functions.

#include "btl/BTL.h"

#if !defined(SGI_CC)
#include <string>
#else
#include <mstring.h>
#endif

#include "btl/Vector.h"
#include "btl/NumericUtilities.h"
#include <stdlib.h>
#include <iomanip.h>

//.............................................................................

unsigned int TestResults(const string& testName, const unsigned int& error)
{
    cout << ">>>>> ERROR = " 
    	 << error
    	 << " : "
    	 << testName 
    	 << " <<<<<\n" 
    	 << endl;
    return error;
}

//.............................................................................

BTL_Vector::value_type CreateReal()
{
    // Set different random number seed each time this function is called.
    static unsigned int seed = 0;
    seed++;
    srand(seed);
    return (BTL_Vector::value_type) rand();
}

//.............................................................................

BTL_Vector CreateVector(BTL_Vector::size_type size)
{
    // Set different random number seed each time this function is called.
    static unsigned int seed = 0;
    seed++;
    srand(seed);

    BTL_Vector vec(size);
    for (BTL_Vector::size_type i=0; i<size; i++)
    	vec[i] = (BTL_Vector::value_type) rand();
    return vec;
}

//.............................................................................

void CheckConstructors()
{
    cout<<"\n############################################\n"
    	 << "### Checking BTL_Vector constructors        ####\n"
    	 << "###                                     ####\n"
    	 << "############################################\n\n";


    BTL_Vector v1;	    	    	    // Construct empty BTL_Vector of size 3

    BTL_Vector v2(1.0,2.0,3.0); 	    // Construct 3D BTL_Vector with initialisation

    BTL_Vector v3(3);   	    	    // Construct empty BTL_Vector of size 3

    unsigned int error = v1.IsNotEqualTo(v3);
    TestResults("Constructing zero BTL_Vectors.",error);

    BTL_Vector::value_type arr[3] = {1.0,2.0,3.0};
    
    BTL_Vector v4(arr,3);	    	    // Construct BTL_Vector from an array.

    error = v2.IsNotEqualTo(v4);
    TestResults("Constructing BTL_Vectors with non-zero elements.",error);
    
    BTL_Vector v5 = v4; 	    	    // Copy construct a BTL_Vector

    error = v4.IsNotEqualTo(v5);
    TestResults("The copy constructor.",error);
}

//.............................................................................

void CheckConstAccessFunctions(const BTL_Vector& v)
{
    cout<<"\n############################################\n"
    	 << "### Checking const access functions     ####\n"
    	 << "###                                     ####\n"
    	 << "############################################\n\n";

    typedef BTL_Vector::const_iterator iterator;
    typedef BTL_Vector::size_type size_type;
    typedef BTL_Vector::value_type value_type;

    unsigned int error, maxError=0;
    size_type indx0 = 0, indx1 = 1;    
    for (iterator i=v.begin(); i!=v.end(); i++, indx0++, indx1++)
    {
    	error = BTL_NumericUtilities<value_type>::AreNotEqual(*i,v[indx0]);
    	if (error > maxError) maxError=error;
    	error = BTL_NumericUtilities<value_type>::AreNotEqual(*i,v(indx1));
    	if (error > maxError) maxError=error;
    }    	

    TestResults("const versions of BTL_Vector iterators and () and [] operators.",
                maxError);
}

//.............................................................................

void CheckOtherAccessFunctions(const BTL_Vector& v)
{
    cout<<"\n############################################\n"
    	 << "### Checking other access functions     ####\n"
    	 << "###                                     ####\n"
    	 << "############################################\n\n";

    typedef BTL_Vector::const_iterator const_iterator;
    typedef BTL_Vector::iterator iterator;
    typedef BTL_Vector::size_type size_type;

    BTL_Vector vcopy(v.size());    

    size_type indx = 0;
    // extra curly brackets are for the benefit of SGI CC version 7.1
    {for (const_iterator i=v.begin(); i!=v.end(); i++, indx++)
     	 vcopy[indx] = *i;}

    unsigned int error = vcopy.IsNotEqualTo(v);
    TestResults("Non - const version the [] operator.", error);

    indx = 1;
    // extra curly brackets are for the benefit of SGI CC version 7.1
    {for (const_iterator i=v.begin(); i!=v.end(); i++, indx++)
     	 vcopy(indx) = *i;}

    error = vcopy.IsNotEqualTo(v);
    TestResults("Non - const version the () operator.", error);
    
    indx = 0;
    for (iterator i=vcopy.begin(); i!=vcopy.end(); i++, indx++)
     	 *i =  v[indx];

    error = vcopy.IsNotEqualTo(v);
    TestResults("Non - const version the BTL_Vector iterator.", error);
}

//.............................................................................

void CheckAddAndSubtraction(const BTL_Vector& v1, const BTL_Vector& v2,
                            BTL_Vector::value_type &shift)
{
    cout<<"\n############################################\n"
    	 << "### Checking addition and subtraction   ####\n"
    	 << "###          operators                  ####\n"
    	 << "############################################\n\n";
    	 
    BTL_Vector v3 = v1 + v2;
    v3 += v2;
    v3 -= v2;
    v3 = v3 - v1;

    unsigned int error = v3.IsNotEqualTo(v2);
    TestResults("Adding and subtracting BTL_Vectors.", error);
    
    BTL_Vector v4 = v1;
    v4 += shift;
    v4 = v4 + shift;
    v4 -= shift;
    v4 = v4 - shift;

    error = v4.IsNotEqualTo(v1);
    TestResults("Adding and subtracting scalars to/from BTL_Vectors.", error);
}


//.............................................................................

void CheckDotProduct(const BTL_Vector& v1, const BTL_Vector& v2, const BTL_Vector& v3)
{
    cout<<"\n############################################\n"
    	 << "### Checking dot/scalar product         ####\n"
    	 << "###       (the * operator)              ####\n"
    	 << "############################################\n\n";

    typedef BTL_Vector::value_type value_type;

    value_type lhs = v1 * (v2 + v3);
    value_type rhs = v1 * v2 + v1 * v3;

    unsigned int error = BTL_NumericUtilities<value_type>::AreNotEqual(lhs,rhs);
    TestResults("v1 * (v2 + v3) == v1 * v2 + v1 * v3 ?", error);

    lhs = v1 * v1;
    rhs = v1.Modulus() * v1.Modulus();

    error = BTL_NumericUtilities<value_type>::AreNotEqual(lhs,rhs);
    TestResults("v1 * v1 == v1.Modulus() * v1.Modulus() ?", error);

    BTL_Vector v4 = v1/v1.Modulus();
    lhs = v4 * v4;
    rhs = 1.0;

    error = BTL_NumericUtilities<value_type>::AreNotEqual(lhs,rhs);
    TestResults("v1/v1.Modulus() * v1/v1.Modulus() == 1.0 ?", error);

    lhs = v1 * v2;
    rhs = v2 * v1;

    error = BTL_NumericUtilities<value_type>::AreNotEqual(lhs,rhs);
    TestResults("v1 * v2 == v2 * v1 ?", error);
}

//.............................................................................

void CheckCrossProduct(const BTL_Vector& v1, const BTL_Vector& v2, 
                               const BTL_Vector& v3)
{
    cout<<"\n############################################\n"
    	 << "### Checking BTL_Vector (cross) product     ####\n"
    	 << "###        (the % operator)             ####\n"
    	 << "############################################\n\n";

    BTL_Vector lhs = v1 % v2, 
           rhs = -(v2 % v1);
    unsigned int error = lhs.IsNotEqualTo(rhs);
    TestResults("a%b = -(b%a) ?", error);


    lhs = v1 % (v2 + v3);
    rhs = v1 % v2 + v1 % v3;
    
    error = lhs.IsNotEqualTo(rhs);
    TestResults("v1 % (v2 + v3) == v1 % v2 + v1 % v3 ?", error);

    typedef BTL_Vector::value_type value_type;

    value_type lreal = v1 * (v2 % v3),
               rreal = v2 * (v3 % v1);
    error = BTL_NumericUtilities<value_type>::AreNotEqual(lreal,rreal);

    TestResults("v1 * (v2 % v3) == v2 * (v3 % v1) ?", error);

    lreal = v1 * (v2 % v3);
    rreal = -(v2*(v1 % v3));
    error = BTL_NumericUtilities<value_type>::AreNotEqual(lreal,rreal);
    TestResults("v1 * (v2 % v3) == -(v2 * (v1 % v3)) ?", error);    
}

//.............................................................................

void CheckMultByScalar(const BTL_Vector& v1, const BTL_Vector& v2,
                       const double& s1, const double& s2)
{
    cout<<"\n############################################\n"
    	 << "### Checking multiplication/division    ####\n"
    	 << "###    of BTL_Vector by a scalar            ####\n"
    	 << "############################################\n\n";

    BTL_Vector lhs = s1 * (s2 * v1),
           rhs = s1 * s2 * v1;
           
    unsigned int error = lhs.IsNotEqualTo(rhs);
    TestResults("s1 * (s2 * v1) == s1 * s2 * v1 ?", error);

    lhs = s1 * (s2 * v1);
    rhs = s2 * (s1 * v1);
    error = lhs.IsNotEqualTo(rhs);
    TestResults("s1 * (s2 * v1) == s2 * (s1 * v1) ?", error);

    lhs = (s1 + s2) * v1;
    rhs = s1 * v1 + s2 * v1;
    error = lhs.IsNotEqualTo(rhs);
    TestResults("(s1 + s2) * v1 == s1 * v1 + s2 * v1 ?", error);

    lhs = s1 * (v1 + v2);
    rhs = s1 * v1 + s1 * v2;
    error = lhs.IsNotEqualTo(rhs);
    TestResults("s1 * (v1 + v2) == s1 * v1 + s1 * v2 ?", error);

    typedef BTL_Vector::value_type value_type;
    value_type lreal = (s1 * v1) * (s2 * v2),
               rreal = s1 * s2 * (v1 * v2);
    error = BTL_NumericUtilities<value_type>::AreNotEqual(lreal,rreal);
    TestResults("(s1 * v1) * (s2 * v2) == s1 * s2 * (v1 * v2) ?", error);

    lhs = (s1 * v1) % (s2 * v2);
    rhs = s1 * s2 * (v1 % v2);
    error = lhs.IsNotEqualTo(rhs);
    TestResults("(s1 * v1) % (s2 * v2) == s1 * s2 * (v1 % v2) ?", error);
}

//.............................................................................

void CheckModulusAndDistance(const BTL_Vector& v1, const BTL_Vector& v2)
{
    cout<<"\n############################################\n"
    	 << "### Checking Modulus() and Distance()   ####\n"
    	 << "###                                     ####\n"
    	 << "############################################\n\n";
    	 
    typedef BTL_Vector::value_type value_type;
    
    value_type dist = v1.Distance(v2);
    BTL_Vector v3 = v2 - v1;
    value_type mod = v3.Modulus();
    
    unsigned int error = BTL_NumericUtilities<value_type>::AreNotEqual(mod,dist);
    TestResults("v1.Distance(v2) == (v2 - v1).Modulus() ?", error);    
}

//.............................................................................
 
// int main(int argc, char* argv[])
// {
//     // check that 1 file name has been supplied
//     if (argc != 2)
//     {
//         cerr << "Usage: MolTest <input file name>" 
//         << endl;
//         exit(1);
//     }
//     
//     char* inputFile = argv[1];
int main()
{
    BTL_Vector vec1 = CreateVector(10),
     	   vec2 = CreateVector(10),
     	   vec3 = CreateVector(10),

     	   vec4 = CreateVector(3),
     	   vec5 = CreateVector(3),
     	   vec6 = CreateVector(3);

    double s1 = CreateReal(), s2 = CreateReal();
    
    CheckConstructors();
    CheckConstAccessFunctions(vec1);
    CheckOtherAccessFunctions(vec1);
    CheckAddAndSubtraction(vec1,vec2,s1);
    CheckDotProduct(vec1,vec2,vec3);
    CheckCrossProduct(vec4,vec5,vec6);
    CheckMultByScalar(vec4,vec5,s1,s2);
    CheckModulusAndDistance(vec1,vec2);
    
    return 0;
}
