#include "Molecule.h"

// Output operator (print function) for Displayable class
//

unsigned int Displayable::nextId = 1;

ostream& operator<<(ostream& os, const Displayable& d)
{
    os << "Displayable  \n";
    os << *d.actor << '\n';
    return os;
}

//VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV
//||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA

// Create empty window, renderer and interactor objects and set background to
// cream.

DisplayWindow::DisplayWindow() 
{
    renWin = rm.MakeRenderWindow(); 	    	    // Create window object
    ren = renWin->MakeRenderer();   	    	    // Create render object
    	ren->SetBackground(1,1,0.5);	    	    // Set background colour
    iren = renWin->MakeRenderWindowInteractor();    // Create interactor
    iren->SetUserMethod(ToggleResolution, (void *) this);     
    	    	    	    	    	    	    // Toggle resolution on 
                                                    // pressing <u>
}
        
// Function to write out scene to VRML file. Call before starting interactor
// 

void
DisplayWindow::Export(char *fileName)
{
    vtkVRMLExporter exporter;	    	// Create exporter object
    exporter.SetInput(renWin);	    	// Tell it what to export
    exporter.SetFilename(fileName); 	// Tell it where to export to
    exporter.Write();	    	    	// Tell it to do its stuff
}

// Render scene then start the interactor
//
void
DisplayWindow::Start()
{
    renWin->Render();				    // Render scene

    // Output options available to user when interactor starts
    // N.B. must flush output buffer so that message appears when expected
    //
    cout << "Mouse buttons:\n <1> rotate\n <2> pan\n <3> zoom\n\n"
         << "Keyboard:\n <w> wireframe\n <s> surface\n <u> toggle resolution\n "
         << "<e> exit\n"
         << endl;
    
    iren->Start();				    // Start interactor
}

// Toggle the resolution of each of the displayable objects displayed 
// within this window. Note this function is friend to the DisplayWindow class
// and not within its scope (i.e. the DisplayWindow:: prefix is not appropriate)
//
void ToggleResolution(void *v)
{
    DisplayWindow *d = (DisplayWindow*) v;
    // Use typedef to type name more readable
    //
    typedef DisplayWindow::Container::iterator iterator;
    
    // Get iterators to the beginning and one past the end of the
    // vector of displaybles
    // 
    iterator i = d->displayables.begin(),
	   end = d->displayables.end();
    
    // Loop through each displayable in a window and toggle its resolution
    //        
    while (i != end)
    {
	Displayable* displayable = *i;      // dereference iterator
	displayable->ToggleResolution();    // call virtual method
	i++;				    // move iterator on to next 
    }					    // item in the vector

    // Re-render scene so that changes take effect
    //
    d->renWin->Render();
}

// Output operator (print function) for DisplayWindow class
//
ostream& operator<<(ostream& os, const DisplayWindow& w)
{
    os << *w.iren << "\n\n";
    return os;
}

//VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV
//||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA

// Initialise atom resolution values - same for each atom

int Atom::highRes = 8;
int Atom::lowRes = 3;

void						    
Atom::GetElementColours(char a1, char a2, float& r, float& g, float& b) 
{
    char A1 = toupper(a1);
    switch(A1)
    {
    case 'C' :
	r=0; g=1; b=0; break;
    case 'N' :
	r=0; g=0; b=1; break;
    case 'O' :
	r=1; g=0; b=0; break;
    default  :
    	r=0; g=0; b=0;
    }
}

// Create atom actor of radius 0.3, white in colour, and
// low resolution

void
Atom::Initialise()
{				
    bestResolution = false;		    	// Start with low resolution
    sphere = new vtkSphereSource;   	    	// Create new sphere object
    	sphere->SetRadius(0.3);
    	sphere->SetThetaResolution(lowRes);
    	sphere->SetPhiResolution(lowRes);
    map = new vtkPolyMapper;
    	map->SetInput(sphere->GetOutput());
    	
    actor->SetMapper(map);
//    actor->GetProperty()->SetFlat();
}

// Create default atom at origin
//

Atom::Atom()
{
    Initialise();
}

// Create at a specified location and with a particular colour
//

Atom::Atom(const float x, const float y, const float z, 
     const float r, const float g, const float b)
{
    Initialise();
    position[0] = x;
    position[1] = y;
    position[2] = z;
    actor->GetProperty()->SetColor(r,g,b);	// sphere color blue
    actor->SetPosition(x,y,z);
}

// Copy constructor
//

Atom::Atom(const Atom& a)
{
    sphere = new vtkSphereSource(*(a.sphere));
    map = new vtkPolyMapper(*(a.map));
    element[0] = a.element[0];
    element[1] = a.element[1];
    position[0] = a.position[0];
    position[1] = a.position[1];
    position[2] = a.position[2];
    bestResolution = a.bestResolution;
}
     
// Assignment operator
//
void
Atom::operator=(const Atom& a)
{
    sphere = new vtkSphereSource(*(a.sphere));
    map = new vtkPolyMapper(*(a.map));
    element[0] = a.element[0];
    element[1] = a.element[1];
    position[0] = a.position[0];
    position[1] = a.position[1];
    position[2] = a.position[2];
    bestResolution = a.bestResolution;
}    

// Toggle resolution (the number of triangles used to create shape)

void
Atom::ToggleResolution()
{
    bestResolution = !bestResolution;
    if (bestResolution)
    {
    	sphere->SetThetaResolution(highRes);
    	sphere->SetPhiResolution(highRes); 
    }
    else
    { 
    	sphere->SetThetaResolution(lowRes);
    	sphere->SetPhiResolution(lowRes); 
    }
}    
	    	
// Output operator (print function) for Atom class
//
ostream& operator<<(ostream& os, const Atom& atom)
{
    os << "\nAtom " << atom.GetId();
    return os;
}


//VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV
//||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA

// Initialise bond resolution values - same for each bond

int Bond::highRes = 8;
int Bond::lowRes = 3;

// Create bond actor of radius 0.1, green in colour, and
// low resolution

void
Bond::Initialise()
{		
    bestResolution = false;		
    cylinder = new vtkCylinderSource;
    	cylinder->SetRadius(0.1);
    	cylinder->SetResolution(lowRes);
    	cylinder->CappingOff();

    // map cylinder to graphics library
    //
    map = new vtkPolyMapper;
    	map->SetInput(cylinder->GetOutput());
    
    // Make actor into cylinder
    //
    actor->SetMapper(map);
    actor->GetProperty()->SetColor(0,0.25,0);
//    actor->GetProperty()->SetFlat();
}

// Create bond between two given atoms
//

Bond::Bond(Atom *atom1, Atom *atom2, float bondLength)
{
    Initialise();

    float a1[3], a2[3];
    
    a1[0] = atom1->GetX();
    a1[1] = atom1->GetY();
    a1[2] = atom1->GetZ();
    
    a2[0] = atom2->GetX();
    a2[1] = atom2->GetY();
    a2[2] = atom2->GetZ();
    
    cylinder->SetHeight(bondLength);
    
    // The cylinder produced thus far by vtk is the correct length but lies on
    // the y-axis with its centre at the origin. A rotation and a translation
    // are needed to bring it to the correct orientation and position. The
    // rotation will be about the origin about the axis along the normal to
    // the plane formed by a vector on the y-axis and the vector between the
    // two input positions (positions of the bonded atoms). The cross product
    // of these two vectors gives us this normal. The angle of rotation is the
    // angle between these two vectors and calculated from the dot product of
    // these two vectors.
    
    // get axis of rotation (normal to plane formed by a vector along the
    // bond and a vector lieing on the y-axis
    //
    // Find coordinates of vector lieing along the bond ( vec2 - vec1 )
    //
    float xbond = a2[0]-a1[0],
    	  ybond = a2[1]-a1[1],
    	  zbond = a2[2]-a1[2];
    
    // calculate cross product of vectors forming plane
    //
    // x = b*r - c*q
    // y = c*p - a*r
    // z = a*q - b*p
    //
    // Since the second vector lies on the y-axis then most terms in the above
    // equations are zero.
    //
    float x_axis = -zbond*bondLength,
    	  y_axis = 0,
    	  z_axis = xbond*bondLength;

    // calculate angle between vectors from the dot product
    //
    // prod = a*p + b*q + a*r
    // prod = modulus(v1)*modulus(v2)*cos(theta)
    //
    // Again the equations are simplified by one vector lieing on the y-axis
    // (p=0 & r=0). Both vectors having modulus = length. M_PI = pi.
    // The minus sign is necessary to ensure a rotation in correct direction.
    //
    float angle = -acos(ybond/bondLength)*180/M_PI;
        	
    // Use actor's funtions to do transformations
    //
    // Translate so bottom end of cylinder lies at origin
    //
    actor->AddPosition(0,-bondLength/2,0);
    	
    // Apply rotation
    //
    actor->RotateWXYZ(angle,x_axis,y_axis,z_axis);
        
    // Translate to position between the two atoms
    //
    actor->SetPosition(a1[0]+(a2[0]-a1[0])/2,
                       a1[1]+(a2[1]-a1[1])/2,
                       a1[2]+(a2[2]-a1[2])/2);
                      
}

// Toggle resolution (the number of triangles used to create shape)

void
Bond::ToggleResolution()
{
    bestResolution = !bestResolution;
    if (bestResolution)
    	cylinder->SetResolution(highRes);
    else 
    	cylinder->SetResolution(lowRes);
}    	    	
    
// Output operator (print function) for bond class
//
ostream& operator<<(ostream& os, const Bond& bond)
{
    // Print object type and some details about this object
    //
    os << "\nBond " << bond.GetId();
    return os;
}

Molecule::Molecule(BTL_XYZProcessor& p)
{
    BTL_XYZProcessor::CoordStore &coords = p.Coords();
    BTL_XYZProcessor::ElementStore &elements = p.Elements();
    
    int nAtoms = coords.size()/3;
    for (int i=0, j=0, k=0; i<nAtoms; i++)
    {
    	Atom *atom = new Atom();
	
	char a1 = elements[j++], a2 = elements[j++];
    	atom->SetElement(a1,a2);
	
	float x = coords[k++], y = coords[k++], z = coords[k++];
    	atom->SetPosition(x,y,z);
	
    	atomsAndBonds.AddVertex(atom);
    }
    CreateBonds();
}

Molecule::~Molecule()
{
    AtomsAndBonds::iterator vi = atomsAndBonds.begin(),
    	    	    	  vend = atomsAndBonds.end();
    while(vi != vend)
    {
    	delete *vi++;
    }
    
    AtomsAndBonds::EdgeDataIterator ei = atomsAndBonds.BeginEdgeData(),
    	    	    	          eend = atomsAndBonds.EndEdgeData();
    while(ei != eend)
    {
    	delete *ei++;
    }
}    
    
void Molecule::CreateBonds()
{
    if (atomsAndBonds.size() < 2) return;

    const float distCutoff = 1.7, cutSqr = distCutoff*distCutoff;
    float dx, dy, dz, distSqr;
    
    typedef AtomsAndBonds::Vertex Vertex;
    AtomsAndBonds::VertexIterator i,j,lastButOne;
    lastButOne=atomsAndBonds.end_vertex();
    lastButOne--;
    for (i=atomsAndBonds.begin_vertex(); i!=lastButOne; i++)    
        for (j=i, j++; j!=atomsAndBonds.end_vertex(); j++)
        {
	    const Vertex& v1 = *i;
	    const Vertex& v2 = *j;
            Atom *atom1 = *(v1.GetIterator());
            Atom *atom2 = *(v2.GetIterator());
            
            dx = atom1->position[0] - atom2->position[0];
            dy = atom1->position[1] - atom2->position[1];
            dz = atom1->position[2] - atom2->position[2];
          
            distSqr = dx*dx + dy*dy + dz*dz;
            
            if (distSqr < cutSqr)
	    {
	    	float dist = fsqrt(distSqr);
		Bond *bond = new Bond(atom1,atom2,dist);
                atomsAndBonds.AddEdge(i,j,bond);
	    }
        }
}

void Molecule::AddToWindow(DisplayWindow& win)
{
    AtomsAndBonds::iterator vi = atomsAndBonds.begin(),
    	    	    	  vend = atomsAndBonds.end();
    while(vi != vend)
    {
    	win.AddDisplayable(*vi++);
    }
    
    AtomsAndBonds::EdgeDataIterator ei = atomsAndBonds.BeginEdgeData(),
    	    	    	          eend = atomsAndBonds.EndEdgeData();
    while(ei != eend)
    {
    	win.AddDisplayable(*ei++);
    }
}    
    

ostream& operator<<(ostream& os, const Molecule& m)
{
    os << m.atomsAndBonds;
    return os;
}   
